import random

from axelrod import Actions, Player, random_choice


class AverageCopier(Player):
    """
    The player will cooperate with probability p if the opponent's cooperation ratio is p.
    Starts with random decision.
    """

    name = 'Average Copier'
    classifier = {
        'memory_depth': float('inf'),  # Long memory
        'stochastic': True,
        'makes_use_of': set(),
        'inspects_source': False,
        'manipulates_source': False,
        'manipulates_state': False
    }

    @staticmethod
    def strategy(opponent):
        if len(opponent.history) == 0:
            # Randomly picks a strategy (not affected by history).
            return random_choice(0.5)
        p = opponent.cooperations // len(opponent.history)
        return random_choice(p)


class NiceAverageCopier(Player):
    """Same as Average Copier, but always starts by cooperating."""

    name = 'Nice Average Copier'
    classifier = {
        'memory_depth': float('inf'),  # Long memory
        'stochastic': True,
        'makes_use_of': set(),
        'inspects_source': False,
        'manipulates_source': False,
        'manipulates_state': False
    }

    @staticmethod
    def strategy(opponent):
        if len(opponent.history) == 0:
            return Actions.C
        p = opponent.cooperations // len(opponent.history)
        return random_choice(p)



import csv
from . import eigen

from numpy import mean, median, std

try:
    # Python 2
    from StringIO import StringIO
except ImportError:
    # Python 3
    from io import StringIO


class ResultSet(object):
    """A class to hold the results of a tournament."""

    def __init__(self, players, matches, with_morality=True):
        """
        Parameters
        ----------
            players : list
                a list of player objects.
            matches : list
                a list of dictionaries mapping tuples of player indices to
                completed matches (1 for each repetition)
            with_morality : bool
                a flag to determine whether morality metrics should be
                calculated.
        """
        self.players = players
        self.nplayers = len(players)
        self.matches = matches
        self.nrepetitions = len(matches)

        # Calculate all attributes:
        self.wins = self.build_wins()
        self.match_lengths = self.build_match_lengths()

        self.scores = self.build_scores()
        self.normalised_scores = self.build_normalised_scores()
        self.ranking = self.build_ranking()
        self.ranked_names = self.build_ranked_names()
        self.payoffs = self.build_payoffs()
        self.payoff_matrix = self.build_payoff_matrix()
        self.payoff_stddevs = self.build_payoff_stddevs()
        self.score_diffs = self.build_score_diffs()
        self.payoff_diffs_means = self.build_payoff_diffs_means()

        if with_morality:
            self.cooperation = self.build_cooperation()
            self.normalised_cooperation = self.build_normalised_cooperation()
            self.vengeful_cooperation = self.build_vengeful_cooperation()
            self.cooperating_rating = self.build_cooperating_rating()
            self.good_partner_matrix = self.build_good_partner_matrix()
            self.good_partner_rating = self.build_good_partner_rating()
            self.eigenmoses_rating = self.build_eigenmoses_rating()
            self.eigenjesus_rating = self.build_eigenjesus_rating()

    @property
    def _null_results_matrix(self):
        """
        Returns:
        --------
            A null matrix (i.e. fully populated with zero values) using
            lists of the form required for the results dictionary.

            i.e. one row per player, containing one element per opponent (in
            order of player index) which lists values for each repetition.
        """
        plist = list(range(self.nplayers))
        replist = list(range(self.nrepetitions))
        return [[[0 for j in plist] for i in plist] for r in replist]

    def build_match_lengths(self):
        """
        Returns:
        --------
            The match lengths. List of the form:

            [ML1, ML2, ML3..., MLn]

            Where n is the number of repetitions and MLi is a list of the form:

            [Pli1, PLi2, Pli3, ..., Plim]

            Where m is the number of players and Plij is of the form:

            [aij1, aij2, aij3, ..., aijk]

            Where k is the number of players and aijk is the length of the match
            between player j and k in repetition i.
        """
        match_lengths = self._null_results_matrix

        for rep in range(self.nrepetitions):

            for player_pair_index, match in self.matches[rep].items():
                i, j = player_pair_index
                match_lengths[rep][i][j] = len(match)

                if i != j:  # Match lengths are symmetric
                    match_lengths[rep][j][i] = len(match)

        return match_lengths

    def build_scores(self):
        """
        Returns:
        --------
            The total scores per player for each repetition lengths.
            List of the form:

            [ML1, ML2, ML3..., MLn]

            Where n is the number of players and MLi is a list of the form:

            [pi1, pi2, pi3, ..., pim]

            Where m is the number of repetitions and pij is the total score
            obtained by each player in repetition j.

            In Axelrod's original tournament, there were no self-interactions
            (e.g. player 1 versus player 1) and so these are also ignored.
        """
        scores = [[0 for rep in range(self.nrepetitions)] for _ in
                  range(self.nplayers)]

        for rep, matches_dict in enumerate(self.matches):
            for index_pair, match in matches_dict.items():
                if index_pair[0] != index_pair[1]: # Ignoring self interactions
                    for player in range(2):
                        player_index = index_pair[player]
                        player_score = match.final_score()[player]
                        scores[player_index][rep] += player_score

        return scores

    def build_ranked_names(self):
        """
        Returns:
        --------
            Returns the ranked names. A list of names as calculated by
            self.ranking.
        """
        return [str(self.players[i]) for i in self.ranking]

    def build_wins(self):
        """
        Returns:
        --------

            The total wins per player for each repetition lengths.
            List of the form:

            [ML1, ML2, ML3..., MLn]

            Where n is the number of players and MLi is a list of the form:

            [pi1, pi2, pi3, ..., pim]

            Where m is the number of repetitions and pij is the total wins
            obtained by each player in repetition j.

            In Axelrod's original tournament, there were no self-interactions
            (e.g. player 1 versus player 1) and so these are also ignored.
        """
        wins = [[0 for rep in range(self.nrepetitions)] for _ in
                range(self.nplayers)]

        for rep, matches_dict in enumerate(self.matches):
            for index_pair, match in matches_dict.items():
                if index_pair[0] != index_pair[1]:  # Ignore self interactions
                    for player in range(2):
                        player_index = index_pair[player]

                        if match.players[player] == match.winner():
                            wins[player_index][rep] += 1

        return wins

    def build_normalised_scores(self):
        """
        Returns:
        --------

            The total mean scores per turn per layer for each repetition
            lengths.  List of the form:

            [ML1, ML2, ML3..., MLn]

            Where n is the number of players and MLi is a list of the form:

            [pi1, pi2, pi3, ..., pim]

            Where m is the number of repetitions and pij is the mean scores per
            turn obtained by each player in repetition j.

            In Axelrod's original tournament, there were no self-interactions
            (e.g. player 1 versus player 1) and so these are also ignored.
        """
        normalised_scores = [
            [[] for rep in range(self.nrepetitions)] for _ in
            range(self.nplayers)]

        # Getting list of all per turn scores for each player for each rep
        for rep, matches_dict in enumerate(self.matches):
            for index_pair, match in matches_dict.items():
                if index_pair[0] != index_pair[1]:  # Ignore self interactions
                    for player in range(2):
                        player_index = index_pair[player]
                        score_per_turn = match.final_score_per_turn()[player]
                        normalised_scores[player_index][rep].append(score_per_turn)

        # Obtaining mean scores and overwriting corresponding entry in
        # normalised scores
        for i, rep in enumerate(normalised_scores):
            for j, player_scores in enumerate(rep):
                normalised_scores[i][j] = mean(player_scores)

        return normalised_scores

    def build_ranking(self):
        """
        Returns:
        --------

            The ranking. List of the form:

            [R1, R2, R3..., Rn]

            Where n is the number of players and Rj is the rank of the jth player
            (based on median normalised score).
        """
        return sorted(range(self.nplayers),
                      key=lambda i: -median(self.normalised_scores[i]))

    def build_payoffs(self):
        """
        Returns:
        --------

            The list of per turn payoffs.
            List of the form:

            [ML1, ML2, ML3..., MLn]

            Where n is the number of players and MLi is a list of the form:

            [pi1, pi2, pi3, ..., pim]

            Where m is the number of players and pij is a list of the form:

            [uij1, uij2, ..., uijk]

            Where k is the number of repetitions and uijk is the list of utilities
            obtained by player i against player j in each repetition.
        """
        plist = list(range(self.nplayers))
        payoffs = [[[] for j in plist] for i in plist]

        for i in plist:
            for j in plist:
                utilities = []
                for rep in self.matches:

                    if (i, j) in rep:
                        match = rep[(i, j)]
                        utilities.append(match.final_score_per_turn()[0])
                    if (j, i) in rep:
                        match = rep[(j, i)]
                        utilities.append(match.final_score_per_turn()[1])

                    payoffs[i][j] = utilities
        return payoffs

    def build_payoff_matrix(self):
        """
        Returns:
        --------
            The mean of per turn payoffs.
            List of the form:

            [ML1, ML2, ML3..., MLn]

            Where n is the number of players and MLi is a list of the form:

            [pi1, pi2, pi3, ..., pim]

            Where m is the number of players and pij is a list of the form:

            [uij1, uij2, ..., uijk]

            Where k is the number of repetitions and u is the mean utility (over
            all repetitions) obtained by player i against player j.
        """
        plist = list(range(self.nplayers))
        payoff_matrix = [[[0] for j in plist] for i in plist]

        for i in plist:
            for j in plist:
                utilities = self.payoffs[i][j]

                if utilities:
                    payoff_matrix[i][j] = mean(utilities)
                else:
                    payoff_matrix[i][j] = 0

        return payoff_matrix

    def build_payoff_stddevs(self):
        """
        Returns:
        --------

            The mean of per turn payoffs.
            List of the form:

            [ML1, ML2, ML3..., MLn]

            Where n is the number of players and MLi is a list of the form:

            [pi1, pi2, pi3, ..., pim]

            Where m is the number of players and pij is a list of the form:

            [uij1, uij2, ..., uijk]

            Where k is the number of repetitions and u is the standard
            deviation of the utility (over all repetitions) obtained by player
            i against player j.
        """
        plist = list(range(self.nplayers))
        payoff_stddevs = [[[0] for j in plist] for i in plist]

        for i in plist:
            for j in plist:
                utilities = self.payoffs[i][j]

                if utilities:
                    payoff_stddevs[i][j] = std(utilities)
                else:
                    payoff_stddevs[i][j] = 0

        return payoff_stddevs

    def build_score_diffs(self):
        """
        Returns:
        --------

            Returns the score differences between players.
            List of the form:

            [ML1, ML2, ML3..., MLn]

            Where n is the number of players and MLi is a list of the form:

            [pi1, pi2, pi3, ..., pim]

            Where m is the number of players and pij is a list of the form:

            [uij1, uij2, ..., uijk]

            Where k is the number of repetitions and uijm is the difference of the
            scores per turn between player i and j in repetition m.
        """
        plist = list(range(self.nplayers))
        score_diffs = [[[0] * self.nrepetitions for j in plist] for i in plist]

        for i in plist:
            for j in plist:
                for r, rep in enumerate(self.matches):
                    if (i, j) in rep:
                        scores = rep[(i, j)].final_score_per_turn()
                        diff = (scores[0] - scores[1])
                        score_diffs[i][j][r] = diff
                    if (j, i) in rep:
                        scores = rep[(j, i)].final_score_per_turn()
                        diff = (scores[1] - scores[0])
                        score_diffs[i][j][r] = diff
        return score_diffs

    def build_payoff_diffs_means(self):
        """
        Returns:
        --------

            The score differences between players.
            List of the form:

            [ML1, ML2, ML3..., MLn]

            Where n is the number of players and MLi is a list of the form:

            [pi1, pi2, pi3, ..., pim]

            Where pij is the mean difference of the
            scores per turn between player i and j in repetition m.
        """
        plist = list(range(self.nplayers))
        payoff_diffs_means = [[0 for j in plist] for i in plist]

        for i in plist:
            for j in plist:
                diffs = []
                for rep in self.matches:
                    if (i, j) in rep:
                        scores = rep[(i, j)].final_score_per_turn()
                        diffs.append(scores[0] - scores[1])
                    if (j, i) in rep:
                        scores = rep[(j, i)].final_score_per_turn()
                        diffs.append(scores[1] - scores[0])
                if diffs:
                    payoff_diffs_means[i][j] = mean(diffs)
                else:
                    payoff_diffs_means[i][j] = 0
        return payoff_diffs_means

    def build_cooperation(self):
        """
        Returns:
        --------

            The list of cooperation counts.
            List of the form:

            [ML1, ML2, ML3..., MLn]

            Where n is the number of players and MLi is a list of the form:

            [pi1, pi2, pi3, ..., pim]

            Where pij is the total number of cooperations over all repetitions
            played by player i against player j.
        """
        plist = list(range(self.nplayers))
        cooperations = [[0 for j in plist] for i in plist]

        for i in plist:
            for j in plist:
                if i != j:
                    for rep in self.matches:
                        coop_count = 0

                        if (i, j) in rep:
                            match = rep[(i, j)]
                            coop_count = match.cooperation()[0]
                        if (j, i) in rep:
                            match = rep[(j, i)]
                            coop_count = match.cooperation()[1]

                        cooperations[i][j] += coop_count
        return cooperations

    def build_normalised_cooperation(self):
        """
        Returns:
        --------

            The list of per turn cooperation counts.
            List of the form:

            [ML1, ML2, ML3..., MLn]

            Where n is the number of players and MLi is a list of the form:

            [pi1, pi2, pi3, ..., pin]

            Where pij is the mean number of
            cooperations per turn played by player i against player j in each
            repetition.
        """
        plist = list(range(self.nplayers))
        normalised_cooperations = [[0 for j in plist] for i in plist]

        for i in plist:
            for j in plist:
                coop_counts = []
                for rep in self.matches:

                    if (i, j) in rep:
                        match = rep[(i, j)]
                        coop_counts.append(match.normalised_cooperation()[0])

                    if (j, i) in rep:
                        match = rep[(j, i)]
                        coop_counts.append(match.normalised_cooperation()[1])

                    if ((i, j) not in rep) and ((j, i) not in rep):
                        coop_counts.append(0)

                    # Mean over all reps:
                    normalised_cooperations[i][j] = mean(coop_counts)
        return normalised_cooperations

    def build_vengeful_cooperation(self):
        """
        Returns:
        --------

            The vengeful cooperation matrix derived from the
            normalised cooperation matrix:

                Dij = 2(Cij - 0.5)
        """
        return [[2 * (element - 0.5) for element in row]
                for row in self.normalised_cooperation]

    def build_cooperating_rating(self):
        """
        Returns:
        --------

            The list of cooperation counts
            List of the form:

            [ML1, ML2, ML3..., MLn]

            Where n is the number of players and MLi is a list of the form:

            [pi1, pi2, pi3, ..., pim]

            Where pij is the total number of cooperations divided by the total
            number of turns over all repetitions played by player i against
            player j.
        """

        plist = list(range(self.nplayers))
        total_length_v_opponent = [zip(*[rep[playeri] for
                                         rep in self.match_lengths])
                                   for playeri in plist]
        lengths = [[sum(e) for j, e in enumerate(row) if i != j] for i, row in
                   enumerate(total_length_v_opponent)]

        # Max is to deal with edge cases of matches that have no turns
        return [sum(cs) / max(1, float(sum(ls))) for cs, ls
                in zip(self.cooperation, lengths)]

    def build_good_partner_matrix(self):
        """
        Returns:
        --------

            An n by n matrix of good partner ratings for n players i.e. an n by
            n matrix where n is the number of players. Each row (i) and column
            (j) represents an individual player and the value Pij is the sum of
            the number of repetitions where player i cooperated as often or
            more than opponent j.
        """

        plist = list(range(self.nplayers))
        good_partner_matrix = [[0 for j in plist] for i in plist]

        for i in plist:
            for j in plist:
                if i != j:
                    for rep in self.matches:

                        if (i, j) in rep:
                            match = rep[(i, j)]
                            coops = match.cooperation()
                            if coops[0] >= coops[1]:
                                good_partner_matrix[i][j] += 1

                        if (j, i) in rep:
                            match = rep[(j, i)]
                            coops = match.cooperation()
                            if coops[0] <= coops[1]:
                                good_partner_matrix[i][j] += 1

        return good_partner_matrix

    def build_good_partner_rating(self):
        """
        Returns:
        --------

        A list of good partner ratings ordered by player index.
        """
        plist = list(range(self.nplayers))
        good_partner_rating = []

        for playeri in plist:
            total_interactions = 0
            for rep in self.matches:
                total_interactions += len(
                    [pair for pair in rep.keys()
                     if playeri in pair and pair[0] != pair[1]])
            # Max is to deal with edge case of matchs with no turns
            rating = sum(self.good_partner_matrix[playeri]) / max(1, float(total_interactions))
            good_partner_rating.append(rating)

        return good_partner_rating

    def build_eigenjesus_rating(self):
        """
        Returns:
        --------

        The eigenjesus rating as defined in:
        http://www.scottaaronson.com/morality.pdf
        """
        eigenvector, eigenvalue = eigen.principal_eigenvector(
                self.normalised_cooperation)
        return eigenvector.tolist()

    def build_eigenmoses_rating(self):
        """
        Returns:
        --------

        The eigenmoses rating as defined in:
        http://www.scottaaronson.com/morality.pdf
        """
        eigenvector, eigenvalue = eigen.principal_eigenvector(
                self.vengeful_cooperation)
        return eigenvector.tolist()

    def csv(self):
        """
        Returns:
        --------

        The string of the total scores per player (columns) per repetition
        (rows).
        """
        csv_string = StringIO()
        header = ",".join(self.ranked_names) + "\n"
        csv_string.write(header)
        writer = csv.writer(csv_string, lineterminator="\n")
        for irep in range(self.nrepetitions):
            data = [self.normalised_scores[rank][irep]
                    for rank in self.ranking]
            writer.writerow(list(map(str, data)))
        return csv_string.getvalue()



from axelrod import Actions, Player

C, D = Actions.C, Actions.D


class Prober(Player):
    """
    Plays D, C, C initially. Defects forever if opponent cooperated in moves 2
    and 3. Otherwise plays TFT.
    """

    name = 'Prober'
    classifier = {
        'stochastic': False,
        'memory_depth': float('inf'),  # Long memory
        'makes_use_of': set(),
        'inspects_source': False,
        'manipulates_source': False,
        'manipulates_state': False
    }

    def strategy(self, opponent):
        turn = len(self.history)
        if turn == 0:
            return D
        if turn == 1:
            return C
        if turn == 2:
            return C
        if turn > 2:
            if opponent.history[1: 3] == [C, C]:
                return D
            else:
                # TFT
                return D if opponent.history[-1:] == [D] else C


class Prober2(Player):
    """
    Plays D, C, C initially. Cooperates forever if opponent played D then C
    in moves 2 and 3. Otherwise plays TFT.
    """

    name = 'Prober 2'
    classifier = {
        'stochastic': False,
        'memory_depth': float('inf'),  # Long memory
        'makes_use_of': set(),
        'inspects_source': False,
        'manipulates_source': False,
        'manipulates_state': False
    }

    def strategy(self, opponent):
        turn = len(self.history)
        if turn == 0:
            return D
        if turn == 1:
            return C
        if turn == 2:
            return C
        if turn > 2:
            if opponent.history[1: 3] == [D, C]:
                return C
            else:
                # TFT
                return D if opponent.history[-1:] == [D] else C


class Prober3(Player):
    """
    Plays D, C initially. Defects forever if opponent played C in moves 2.
    Otherwise plays TFT.
    """

    name = 'Prober 3'
    classifier = {
        'stochastic': False,
        'memory_depth': float('inf'),  # Long memory
        'makes_use_of': set(),
        'inspects_source': False,
        'manipulates_source': False,
        'manipulates_state': False
    }

    def strategy(self, opponent):
        turn = len(self.history)
        if turn == 0:
            return D
        if turn == 1:
            return C
        if turn > 1:
            if opponent.history[1] == C:
                return D
            else:
                # TFT
                return D if opponent.history[-1:] == [D] else C


class HardProber(Player):
    """
    Plays D, D, C, C initially. Defects forever if opponent cooperated in moves
    2 and 3. Otherwise plays TFT.
    """

    name = 'Hard Prober'
    classifier = {
        'stochastic': False,
        'memory_depth': float('inf'),  # Long memory
        'makes_use_of': set(),
        'inspects_source': False,
        'manipulates_source': False,
        'manipulates_state': False
    }

    def strategy(self, opponent):
        turn = len(self.history)
        if turn == 0:
            return D
        if turn == 1:
            return D
        if turn == 2:
            return C
        if turn == 3:
            return C
        if turn > 3:
            if opponent.history[1: 3] == [C, C]:
                return D
            else:
                # TFT
                return D if opponent.history[-1:] == [D] else C



from axelrod import Actions, Player, init_args

C, D = Actions.C, Actions.D


class Grumpy(Player):
    """A player that defects after a certain level of grumpiness.
    Grumpiness increases when the opponent defects and decreases
    when the opponent co-operates."""

    name = 'Grumpy'
    classifier = {
        'memory_depth': float('inf'),  # Long memory
        'stochastic': False,
        'makes_use_of': set(),
        'inspects_source': False,
        'manipulates_source': False,
        'manipulates_state': False
    }

    @init_args
    def __init__(self, starting_state='Nice', grumpy_threshold=10,
                 nice_threshold=-10):
        """
        Parameters
        ----------
        starting_state, str: 'Nice' or 'Grumpy'
        grumpy_threshold, int
            The threshold of opponent defections - cooperations to become
            grumpy
        nice_threshold, int
            The threshold of opponent defections - cooperations to become
            nice
        """
        super(Grumpy, self).__init__()
        self.history = []
        self.state = starting_state
        self.starting_state = starting_state
        self.grumpy_threshold = grumpy_threshold
        self.nice_threshold = nice_threshold

    def strategy(self, opponent):
        """A player that gets grumpier the more the opposition defects,
        and nicer the more they cooperate.

        Starts off Nice, but becomes grumpy once the grumpiness threshold is hit.
        Won't become nice once that grumpy threshold is hit, but must reach a much lower threshold before it becomes nice again.
        """

        self.grumpiness = opponent.defections - opponent.cooperations

        if self.state == 'Nice':
            if self.grumpiness > self.grumpy_threshold:
                self.state = 'Grumpy'
                return D
            return C

        if self.state == 'Grumpy':
            if self.grumpiness < self.nice_threshold:
                self.state = 'Nice'
                return C
            return D

    def reset(self):
        """Resets score, history and state for the next round of the tournement."""
        Player.reset(self)
        self.state = self.starting_state



from axelrod import Actions

C, D = Actions.C, Actions.D

class Game(object):
    """A class to hold the game matrix and to score a game accordingly."""

    def __init__(self, r=3, s=0, t=5, p=1):
        self.scores = {
            (C, C): (r, r),
            (D, D): (p, p),
            (C, D): (s, t),
            (D, C): (t, s),
        }

    def RPST(self):
        """Return the values in the game matrix in the Press and Dyson notation."""
        R = self.scores[(C, C)][0]
        P = self.scores[(D, D)][0]
        S = self.scores[(C, D)][0]
        T = self.scores[(D, C)][0]
        return (R, P, S, T)

    def score(self, pair):
        """Return the appropriate score for decision pair.

        Returns the appropriate score (as a tuple) from the scores dictionary
        for a given pair of plays (passed in as a tuple).
        e.g. score((C, C)) returns (2, 2)
        """
        return self.scores[pair]

DefaultGame = Game()



import itertools

from axelrod import Actions, Player
from .axelrod_first import Joss
from axelrod._strategy_utils import detect_cycle

class Calculator(Player):
    """
    Plays like (Hard) Joss for the first 20 rounds. If periodic behavior is
    detected, defect forever. Otherwise play TFT.
    """

    name = "Calculator"
    classifier = {
        'memory_depth': float('inf'),
        'stochastic': True,
        'makes_use_of': set(),
        'inspects_source': False,
        'manipulates_source': False,
        'manipulates_state': False
    }

    def __init__(self):
        Player.__init__(self)
        self.joss_instance = Joss()

    def strategy(self, opponent):
        turn = len(self.history)
        if turn == 20:
            self.cycle = detect_cycle(opponent.history)
            return self.extended_strategy(opponent)
        if turn > 20:
            return self.extended_strategy(opponent)
        else:
            play = self.joss_instance.strategy(opponent)
            self.joss_instance.history.append(play)
            return play

    def extended_strategy(self, opponent):
        if self.cycle:
            return Actions.D
        else:
            # TFT
            return Actions.D if opponent.history[-1:] == [Actions.D] else Actions.C

    def reset(self):
        Player.reset(self)
        self.joss_instance = Joss()



import random
from axelrod import Actions

def random_choice(p=0.5):
    """
    Return 'C' with probability `p`, else return 'D'

    Emulates Python's random.choice(['C', 'D']) since it is not consistent
    across Python 2.7 to Python 3.4"""

    r = random.random()
    if r < p:
        return Actions.C
    return Actions.D



# -*- coding: utf-8 -*-
from .game import Game
from axelrod import Actions

C, D = Actions.C, Actions.D


def sparkline(actions, c_symbol=u'█', d_symbol=u' '):
    return u''.join([
        c_symbol if play == 'C' else d_symbol for play in actions])


class Match(object):

    def __init__(self, players, turns, deterministic_cache=None,
                 cache_mutable=True, noise=0):
        """
        Parameters
        ----------
        players : tuple
            A pair of axelrod.Player objects
        turns : integer
                The number of turns per match
        deterministic_cache : dictionary
            A cache of resulting actions for deterministic matches
        cache_mutable : boolean
            Whether the deterministic cache can be updated or not
        noise : float
            The probability that a player's intended action should be flipped
        """
        self.result = []
        self.players = list(players)
        self._classes = (players[0].__class__, players[1].__class__)
        self._turns = turns
        if deterministic_cache is None:
            self._cache = {}
        else:
            self._cache = deterministic_cache
        self._cache_mutable = cache_mutable
        self._noise = noise

    @property
    def _stochastic(self):
        """
        A boolean to show whether a match between two players would be
        stochastic
        """
        return (
            self._noise or
            any(p.classifier['stochastic'] for p in self.players)
            )

    @property
    def _cache_update_required(self):
        """
        A boolean to show whether the deterministic cache should be updated
        """
        return (
            not self._noise and
            self._cache_mutable and not (
                any(p.classifier['stochastic'] for p in self.players)
                )
        )

    def play(self):
        """
        The resulting list of actions from a match between two players.

        This method determines whether the actions list can be obtained from
        the deterministic cache and returns it from there if so. If not, it
        calls the play method for player1 and returns the list from there.

        Returns
        -------
        A list of the form:

        e.g. for a 2 turn match between Cooperator and Defector:

            [(C, C), (C, D)]

        i.e. One entry per turn containing a pair of actions.
        """
        if (self._stochastic or self._classes not in self._cache):
            turn = 0
            for p in self.players:
                p.reset()
            while turn < self._turns:
                turn += 1
                self.players[0].play(self.players[1], self._noise)
            result = list(
                zip(self.players[0].history, self.players[1].history))

            if self._cache_update_required:
                self._cache[self._classes] = result
        else:
            result = self._cache[self._classes]

        self.result = result
        return result

    def scores(self, game=None):
        """Returns the scores of the previous Match plays."""
        if not game:
            game = Game()
        scores = [game.score(plays) for plays in self.result]
        return scores

    def final_score(self, game=None):
        """Returns the final score for a Match"""
        scores = self.scores(game)

        if len(scores) == 0:
            return None

        final_score = tuple(sum([score[playeri] for score in scores])
                            for playeri in [0, 1])
        return final_score

    def final_score_per_turn(self, game=None):
        """Returns the mean score per round for a Match"""
        scores = self.scores(game)

        if len(scores) == 0:
            return None

        final_score_per_turn = tuple(
            sum([score[playeri] for score in scores]) / (float(self._turns))
            for playeri in [0, 1])
        return final_score_per_turn

    def winner(self, game=None):
        """Returns the winner of the Match"""
        scores = self.final_score(game)

        if scores is not None:
            if scores[0] == scores[1]:
                return False  # No winner
            return sorted(self.players,
                          key=lambda x: scores[self.players.index(x)])[-1]
        return None

    def cooperation(self):
        """Returns the count of cooperations by each player"""

        if len(self.result) == 0:
            return None

        cooperation = tuple(sum([play[playeri] == C for play in self.result])
                            for playeri in [0, 1])
        return cooperation

    def normalised_cooperation(self):
        """Returns the count of cooperations by each player per turn"""
        cooperation = self.cooperation()

        if len(self.result) == 0:
            return None

        normalised_cooperation = tuple(
            [c / float(self._turns) for c in cooperation])

        return normalised_cooperation

    def sparklines(self, c_symbol=u'█', d_symbol=u' '):
        return (
            sparkline(self.players[0].history, c_symbol, d_symbol) +
            u'\n' +
            sparkline(self.players[1].history, c_symbol, d_symbol))

    def __len__(self):
        return self._turns



from axelrod import Actions, Player, obey_axelrod
from ._strategies import strategies
from .hunter import DefectorHunter, AlternatorHunter, RandomHunter, MathConstantHunter, CycleHunter, EventualCycleHunter
from .cooperator import Cooperator
from numpy.random import choice

# Needs to be computed manually to prevent circular dependency
ordinary_strategies = [s for s in strategies if obey_axelrod(s)]
C, D = Actions.C, Actions.D

class MetaPlayer(Player):
    """A generic player that has its own team of players."""

    name = "Meta Player"

    team = [Cooperator]
    classifier = {
        'memory_depth': float('inf'),  # Long memory
        'stochastic': False,
        'makes_use_of': set(),
        'inspects_source': False,
        'manipulates_source': False,
        'manipulates_state': False
    }

    def __init__(self):

        super(MetaPlayer, self).__init__()

        # Make sure we don't use any meta players to avoid infinite recursion.
        self.team = [t for t in self.team if not issubclass(t, MetaPlayer)]
        self.nteam = len(self.team)

        # Initiate all the player in out team.
        self.team = [t() for t in self.team]

        # This player inherits the classifiers of its team.
        for key in ['stochastic',
                    'inspects_source',
                    'manipulates_source',
                    'manipulates_state']:
            self.classifier[key] = (any(t.classifier[key] for t in self.team))

        for t in self.team:
            self.classifier['makes_use_of'].update(t.classifier['makes_use_of'])

    def strategy(self, opponent):

        # Make sure the history of all hunters is current.
        for ih in range(len(self.team)):
            self.team[ih].history = self.history

        # Get the results of all our players.
        results = [player.strategy(opponent) for player in self.team]

        # A subclass should just define a way to choose the result based on team results.
        return self.meta_strategy(results, opponent)

    def meta_strategy(self, results, opponent):
        """Determine the meta result based on results of all players."""
        pass

    def reset(self):
        Player.reset(self)
        # Reset each player as well
        for player in self.team:
            player.reset()


class MetaMajority(MetaPlayer):
    """A player who goes by the majority vote of all other non-meta players."""

    name = "Meta Majority"

    def __init__(self, team=None):
        if team:
            self.team = team
        else:
            # Needs to be computed manually to prevent circular dependency
            self.team = ordinary_strategies
        super(MetaMajority, self).__init__()
        self.init_args = (team,)
        self.classifier['memory_depth'] = float('inf')

    @staticmethod
    def meta_strategy(results, opponent):
        if results.count(D) > results.count(C):
            return D
        return C


class MetaMinority(MetaPlayer):
    """A player who goes by the minority vote of all other non-meta players."""

    name = "Meta Minority"

    def __init__(self, team=None):
        if team:
            self.team = team
        else:
            # Needs to be computed manually to prevent circular dependency
            self.team = ordinary_strategies
        super(MetaMinority, self).__init__()
        self.init_args = (team,)
        self.classifier['memory_depth'] = float('inf')

    @staticmethod
    def meta_strategy(results, opponent):
        if results.count(D) < results.count(C):
            return D
        return C


class MetaWinner(MetaPlayer):
    """A player who goes by the strategy of the current winner."""

    name = "Meta Winner"

    def __init__(self, team=None):

        # The default is to use all strategies available, but we need to import the list
        # at runtime, since _strategies import also _this_ module before defining the list.
        if team:
            self.team = team
        else:
            # Needs to be computed manually to prevent circular dependency
            self.team = ordinary_strategies

        super(MetaWinner, self).__init__()
        self.init_args = (team,)

        # For each player, we will keep the history of proposed moves and
        # a running score since the beginning of the game.
        for t in self.team:
            t.proposed_history = []
            t.score = 0

    def strategy(self, opponent):

        # Update the running score for each player, before determining the next move.
        if len(self.history):
            for player in self.team:
                game = self.tournament_attributes["game"]
                last_round = (player.proposed_history[-1], opponent.history[-1])
                s = game.scores[last_round][0]
                player.score += s
        return super(MetaWinner, self).strategy(opponent)

    def meta_strategy(self, results, opponent):

        scores = [pl.score for pl in self.team]
        bestscore = max(scores)
        beststrategies = [i for i, pl in enumerate(self.team) if pl.score == bestscore]
        bestproposals = [results[i] for i in beststrategies]
        bestresult = C if C in bestproposals else D

        # Update each player's proposed history with his proposed result, but always after
        # the new result has been settled based on scores accumulated until now.
        for r, t in zip(results, self.team):
            t.proposed_history.append(r)

        if opponent.defections == 0:
            # Don't poke the bear
            return C

        return bestresult


class MetaHunter(MetaPlayer):
    """A player who uses a selection of hunters."""

    name = "Meta Hunter"
    classifier = {
        'memory_depth': float('inf'),  # Long memory
        'stochastic': False,
        'makes_use_of': set(),
        'inspects_source': False,
        'manipulates_source': False,
        'manipulates_state': False
    }

    def __init__(self):
        # Notice that we don't include the cooperator hunter, because it leads to excessive
        # defection and therefore bad performance against unforgiving strategies. We will stick
        # to hunters that use defections as cues. However, a really tangible benefit comes from
        # combining Random Hunter and Math Constant Hunter, since together they catch strategies
        # that are lightly randomized but still quite constant (the tricky/suspecious ones).
        self.team = [DefectorHunter, AlternatorHunter, RandomHunter, MathConstantHunter, CycleHunter, EventualCycleHunter]

        super(MetaHunter, self).__init__()

    @staticmethod
    def meta_strategy(results, opponent):

        # If any of the hunters smells prey, then defect!
        if D in results:
            return D

        # Tit-for-tat might seem like a better default choice, but in many cases it complicates
        # the heuristics of hunting and creates fale-positives. So go ahead and use it, but only
        # for longer histories.
        if len(opponent.history) > 100:
            return D if opponent.history[-1:] == [D] else C
        else:
            return C


class MetaMajorityMemoryOne(MetaMajority):
    """MetaMajority with the team of Memory One players"""

    name = "Meta Majority Memory One"

    def __init__(self):
        team = [s for s in ordinary_strategies if s().classifier['memory_depth'] <= 1]
        super(MetaMajorityMemoryOne, self).__init__(team=team)
        self.init_args = ()


class MetaWinnerMemoryOne(MetaWinner):
    """MetaWinner with the team of Memory One players"""

    name = "Meta Winner Memory One"

    def __init__(self):
        team = [s for s in ordinary_strategies if s().classifier['memory_depth'] <= 1]
        super(MetaWinnerMemoryOne, self).__init__(team=team)
        self.init_args = ()


class MetaMajorityFiniteMemory(MetaMajority):
    """MetaMajority with the team of Finite Memory Players"""

    name = "Meta Majority Finite Memory"
    def __init__(self):
        team = [s for s in ordinary_strategies if s().classifier['memory_depth']
                < float('inf')]
        super(MetaMajorityFiniteMemory, self).__init__(team=team)
        self.init_args = ()


class MetaWinnerFiniteMemory(MetaWinner):
    """MetaWinner with the team of Finite Memory Players"""

    name = "Meta Winner Finite Memory"
    def __init__(self):
        team = [s for s in ordinary_strategies if s().classifier['memory_depth']
                < float('inf')]
        super(MetaWinnerFiniteMemory, self).__init__(team=team)
        self.init_args = ()


class MetaMajorityLongMemory(MetaMajority):
    """MetaMajority with the team of Long (infinite) Memory Players"""

    name = "Meta Majority Long Memory"
    def __init__(self):
        team = [s for s in ordinary_strategies if s().classifier['memory_depth']
                == float('inf')]
        super(MetaMajorityLongMemory, self).__init__(team=team)
        self.init_args = ()


class MetaWinnerLongMemory(MetaWinner):
    """MetaWinner with the team of Long (infinite) Memory Players"""

    name = "Meta Winner Long Memory"
    def __init__(self):
        team = [s for s in ordinary_strategies if s().classifier['memory_depth']
                == float('inf')]
        super(MetaWinnerLongMemory, self).__init__(team=team)
        self.init_args = ()


class MetaMixer(MetaPlayer):
    """A player who randomly switches between a team of players.
    If no distribution is passed then the player will uniformly choose between
    sub players.

    In essence this is creating a Mixed strategy.

    Parameters

    team : list of strategy classes, optional
        Team of strategies that are to be randomly played
        If none is passed will select the ordinary strategies.
    distribution : list representing a probability distribution, optional
        This gives the distribution from which to select the players.
        If none is passed will select uniformly.
    """

    name = "Meta Mixer"

    def __init__(self, team=None, distribution=None):

        # The default is to use all strategies available, but we need to import the list
        # at runtime, since _strategies import also _this_ module before defining the list.
        if team:
            self.team = team
        else:
            # Needs to be computed manually to prevent circular dependency
            self.team = ordinary_strategies

        self.distribution = distribution

        super(MetaMixer, self).__init__()
        self.init_args = (team, distribution)

    def meta_strategy(self, results, opponent):
        """Using the numpy.random choice function to sample with weights"""
        return choice(results, p=self.distribution)



from axelrod import Actions, Player, random_choice

C, D = Actions.C, Actions.D

class Inverse(Player):
    """A player who defects with a probability that diminishes relative to how
    long ago the opponent defected."""

    name = 'Inverse'
    classifier = {
        'memory_depth': float('inf'),  # Long memory
        'stochastic': True,
        'makes_use_of': set(),
        'inspects_source': False,
        'manipulates_source': False,
        'manipulates_state': False
    }

    @staticmethod
    def strategy(opponent):
        """Looks at opponent history to see if they have defected.

        If so, player defection is inversely proportional to when this occurred.
        """

        index = next((index for index, value in enumerate(opponent.history, start=1) if value == D), None)

        if index is None:
            return C

        return random_choice(1 - 1 / float(abs(index)))



from __future__ import absolute_import
from collections import OrderedDict

import axelrod.strategies

from .tournament_manager import *


class TournamentManagerFactory(object):


    tournaments = OrderedDict([
        ('basic_strategies', axelrod.basic_strategies),
        ('ordinary_strategies',
            axelrod.ordinary_strategies),
        ('cheating_strategies', axelrod.cheating_strategies),
        ('strategies',
            axelrod.ordinary_strategies +
            axelrod.cheating_strategies)
    ])

    @classmethod
    def create_tournament_manager(
            cls,
            output_directory,
            no_ecological,
            rebuild_cache,
            cache_file,
            exclusions,
            processes,
            turns,
            repetitions,
            noise,
            image_format="svg"):

        kwargs = {
            'processes': processes,
            'turns': turns,
            'repetitions': repetitions,
            'noise': noise,
        }

        manager = axelrod.TournamentManager(
            output_directory=output_directory,
            with_ecological=not no_ecological,
            save_cache=rebuild_cache,
            cache_file=cache_file,
            image_format=image_format)

        cls._add_tournaments(manager, exclusions, kwargs)

        return manager

    @classmethod
    def _add_tournaments(cls, manager, exclusions, kwargs):
        for name, strategies in cls._tournaments_dict(exclusions).items():
            players = manager.one_player_per_strategy(strategies)
            manager.add_tournament(
                name=name, players=players, **kwargs)

    @classmethod
    def _tournaments_dict(cls, exclusions=None):
        if exclusions is None:
            exclusions = []

        return OrderedDict([
            (key, value) for
            key, value in cls.tournaments.items() if key not in exclusions
        ])


class ProbEndTournamentManagerFactory(TournamentManagerFactory):

    @classmethod
    def create_tournament_manager(
            cls,
            output_directory,
            no_ecological,
            rebuild_cache,
            cache_file,
            exclusions,
            processes,
            prob_end,
            repetitions,
            noise,
            image_format="svg"):

        kwargs = {
            'processes': processes,
            'prob_end': prob_end,
            'repetitions': repetitions,
            'noise': noise,
        }

        manager = axelrod.ProbEndTournamentManager(
            output_directory=output_directory,
            with_ecological=not no_ecological,
            save_cache=rebuild_cache,
            cache_file=cache_file,
            image_format=image_format)

        cls._add_tournaments(manager, exclusions, kwargs)

        return manager

    @classmethod
    def _tournaments_dict(cls, exclusions=None):
        if exclusions is None:
            exclusions = []

        return OrderedDict([
            (key+'_prob_end', value) for
            key, value in cls.tournaments.items() if key not in exclusions
        ])




import random


class Ecosystem(object):
    """Create an ecosystem based on the payoff matrix from an Axelrod tournament."""

    def __init__(self, results, fitness=None, population=None):

        self.results = results
        self.nplayers = self.results.nplayers
        self.payoff_matrix = self.results.payoff_matrix
        self.payoff_stddevs = self.results.payoff_stddevs

        # Population sizes will be recorded in this nested list, with each internal
        # list containing strategy populations for a given turn. The first list,
        # representing the starting populations, will by default have all equal
        # values, and all population lists will be normalized to one.
        # An initial population vector can also be passed. This will be
        # normalised, but must be of the correct size and have all
        # non-negative values.
        if population:
            if min(population) < 0:
                raise TypeError("Minimum value of population vector must be non-negative")
            elif len(population) != self.nplayers:
                raise TypeError("Population vector must be same size as number of players")
            else:
                norm = float(sum(population))
                self.population_sizes = [[p / norm for p in population]]
        else:
            self.population_sizes = [[1.0 / self.nplayers for i in range(self.nplayers)]]

        # This function is quite arbitrary and probably only influences the kinetics
        # for the current code.
        if fitness:
            self.fitness = fitness
        else:
            self.fitness = lambda p: p

    def reproduce(self, turns):

        for iturn in range(turns):

            plist = list(range(self.nplayers))
            pops = self.population_sizes[-1]

            # The unit payoff for each player in this turn is the sum of the payoffs
            # obtained from playing with all other players, scaled by the size of the
            # opponent's population. Note that we sample the normal distribution
            # based on the payoff matrix and its standard deviations obtained from
            # the iterated PD tournament run previously.
            payoffs = [0 for ip in plist]
            for ip in plist:
                for jp in plist:
                    avg = self.payoff_matrix[ip][jp]
                    dev = self.payoff_stddevs[ip][jp]
                    p = random.normalvariate(avg, dev)
                    payoffs[ip] += p * pops[jp]

            # The fitness should determine how well a strategy reproduces. The new populations
            # should be multiplied by something that is proportional to the fitness, but we are
            # normalizing anyway so just multiply times fitness.
            fitness = [self.fitness(p) for p in payoffs]
            newpops = [p * f for p, f in zip(pops, fitness)]

            # Make sure the new populations are normalized to one.
            norm = sum(newpops)
            newpops = [p / norm for p in newpops]

            self.population_sizes.append(newpops)



from axelrod import Actions, Player

C, D = Actions.C, Actions.D

class Defector(Player):
    """A player who only ever defects."""

    name = 'Defector'
    classifier = {
        'memory_depth': 0,
        'stochastic': False,
        'makes_use_of': set(),
        'inspects_source': False,
        'manipulates_source': False,
        'manipulates_state': False
    }

    @staticmethod
    def strategy(opponent):
        return D


class TrickyDefector(Player):
    """A defector that is trying to be tricky."""

    name = "Tricky Defector"
    classifier = {
        'memory_depth': float('inf'),  # Long memory
        'stochastic': False,
        'makes_use_of': set(),
        'inspects_source': False,
        'manipulates_source': False,
        'manipulates_state': False
    }

    def strategy(self, opponent):
        """Almost always defects, but will try to trick the opponent into cooperating.

        Defect if opponent has cooperated at least once in the past and has defected
        for the last 3 turns in a row.
        """
        if C in opponent.history and opponent.history[-3:] == [D] * 3:
            return C
        return D



from axelrod import Actions, Player, init_args
from axelrod._strategy_utils import thue_morse_generator


class SequencePlayer(Player):
    """Abstract base class for players that use a generated sequence to
    determine their plays."""

    @init_args
    def __init__(self, generator_function, generator_args=()):
        Player.__init__(self)
        # Initialize the sequence generator
        self.generator_function = generator_function
        self.generator_args = generator_args
        self.sequence_generator = self.generator_function(*self.generator_args)

    def meta_strategy(self, value):
        """Determines how to map the sequence value to cooperate or defect.
        By default, treat values like python truth values. Override in child
        classes for alternate behaviors."""
        if value == 0:
            return Actions.D
        else:
            return Actions.C

    def strategy(self, opponent):
        # Iterate through the sequence and apply the meta strategy
        for s in self.sequence_generator:
            return self.meta_strategy(s)

    def reset(self):
        # Be sure to reset the sequence generator
        Player.reset(self)
        self.sequence_generator = self.generator_function(*self.generator_args)


class ThueMorse(SequencePlayer):
    """
    A player who cooperates or defects according to the Thue-Morse sequence.

    The first few terms of the Thue-Morse sequence are:
    0 1 1 0 1 0 0 1 1 0 0 1 0 1 1 0 . . .
    """

    name = 'ThueMorse'
    classifier = {
        'memory_depth': float('inf'),
        'stochastic': False,
        'makes_use_of': set(),
        'inspects_source': False,
        'manipulates_source': False,
        'manipulates_state': False
    }

    @init_args
    def __init__(self):
        SequencePlayer.__init__(self, thue_morse_generator, (0,))


class ThueMorseInverse(ThueMorse):
    """A player who defects or cooperates according to the Thue-Morse sequence
    (Inverse of ThueMorse)."""

    name = 'ThueMorseInverse'
    classifier = {
        'memory_depth': float('inf'),
        'stochastic': False,
        'makes_use_of': set(),
        'inspects_source': False,
        'manipulates_source': False,
        'manipulates_state': False
    }

    @init_args
    def __init__(self):
        SequencePlayer.__init__(self, thue_morse_generator, (0,))

    def meta_strategy(self, value):
        # Switch the default cooperate and defect action on 0 or 1
        if value == 0:
            return Actions.C
        else:
            return Actions.D




"""
Additional strategies from Axelrod's second tournament.
"""

import random

from axelrod import Actions, Player, flip_action, random_choice

C, D = Actions.C, Actions.D


class Champion(Player):
    """
    Strategy submitted to Axelrod's second tournament by Danny Champion.
    """

    name = "Champion"
    classifier = {
        'memory_depth': float('inf'),
        'stochastic': True,
        'makes_use_of': set(["length"]),
        'inspects_source': False,
        'manipulates_source': False,
        'manipulates_state': False
    }

    def strategy(self, opponent):
        current_round = len(self.history)
        expected_length = self.tournament_attributes['length']
        # Cooperate for the first 1/20-th of the game
        if current_round == 0:
            return C
        if current_round < expected_length / 20.:
            return C
        # Mirror partner for the next phase
        if current_round < expected_length * 5 / 40.:
            return opponent.history[-1]
        # Now cooperate unless all of the necessary conditions are true
        defection_prop = float(opponent.defections) / len(opponent.history)
        if opponent.history[-1] == D:
            r = random.random()
            if defection_prop > max(0.4, r):
                return D
        return C


class Eatherley(Player):
    """
    Strategy submitted to Axelrod's second tournament by Graham Eatherley.
    """

    name = "Eatherley"
    classifier = {
        'memory_depth': float('inf'),
        'stochastic': True,
        'makes_use_of': set(),
        'inspects_source': False,
        'manipulates_source': False,
        'manipulates_state': False
    }

    @staticmethod
    def strategy(opponent):
        # Cooperate on the first move
        if not len(opponent.history):
            return C
        # Reciprocate cooperation
        if opponent.history[-1] == C:
            return C
        # Respond to defections with probability equal to opponent's total
        # proportion of defections
        defection_prop = float(opponent.defections) / len(opponent.history)
        return random_choice(1 - defection_prop)


class Tester(Player):
    """
    Submitted to Axelrod's second tournament by David Gladstein.

    Defects on the first move and plays TFT if the opponent ever defects (after
    one apology cooperation round). Otherwise alternate cooperation and defection.
    """

    name = "Tester"
    classifier = {
        'memory_depth': float('inf'),
        'stochastic': False,
        'makes_use_of': set(),
        'inspects_source': False,
        'manipulates_source': False,
        'manipulates_state': False
    }

    def __init__(self):
        Player.__init__(self)
        self.is_TFT = False

    def strategy(self, opponent):
        # Defect on the first move
        if not opponent.history:
            return D
        # Am I TFT?
        if self.is_TFT:
            return D if opponent.history[-1:] == [D] else C
        else:
            # Did opponent defect?
            if opponent.history[-1] == D:
                self.is_TFT = True
                return C
            if len(self.history) in [1, 2]:
                return C
            # Alternate C and D
            return flip_action(self.history[-1])

    def reset(self):
        Player.reset(self)
        self.is_TFT = False



import random
from math import ceil, log

from .match import Match


class TournamentType(object):

    clone_opponents = True

    def __init__(self, players, turns, deterministic_cache):
        """
        A class to represent a type of tournament and build the set of matches
        accordingly.

        Parameters
        ----------
        players : list
            A list of axelrod.Player objects
        turns : integer
            The number of turns per match
        deterministic_cache : dictionary
            A cache of resulting actions for deterministic matches
        """
        self.players = players
        self.turns = turns
        self.deterministic_cache = deterministic_cache
        self.opponents = players

    @property
    def opponents(self):
        return self._opponents

    @opponents.setter
    def opponents(self, players):
        if self.clone_opponents:
            opponents = []
            for player in players:
                opponents.append(player.clone())
        else:
            opponents = players
        self._opponents = opponents

    def build_matches():
        raise NotImplementedError()

    def build_single_match():
        raise NotImplementedError()


class RoundRobin(TournamentType):

    clone_opponents = True

    def build_matches(self, cache_mutable=True, noise=0):
        """
        Create a dictionary of match objects for a round robin tournament.

        Parameters
        ----------
        cache_mutable : boolean
            Whether the deterministic cache should be updated
        noise : float
            The probability that a player's intended action should be flipped

        Returns
        -------
        dictionary
            Mapping a tuple of player index numbers to an axelrod Match object
        """
        matches = {}
        for player1_index in range(len(self.players)):
            for player2_index in range(player1_index, len(self.players)):
                pair = (
                    self.players[player1_index], self.opponents[player2_index])
                match = self.build_single_match(pair, cache_mutable, noise)
                matches[(player1_index, player2_index)] = match
        return matches

    def build_single_match(self, pair, cache_mutable=True, noise=0):
        """Create a single match for a given pair"""
        return Match(pair, self.turns, self.deterministic_cache,
                     cache_mutable, noise)


class ProbEndRoundRobin(RoundRobin):

    clone_opponents = True

    def __init__(self, players, prob_end, deterministic_cache):
        """
        A class to represent a tournament type for which the players do not
        know the length of the Match (to their knowledge it is infinite) but
        that ends with given probability.

        Parameters
        ----------
        players : list
            A list of axelrod.Player objects
        prob_end : float
            The probability that a turn of a Match is the last
        deterministic_cache : dictionary
            A cache of resulting actions for deterministic matches
        """
        super(ProbEndRoundRobin, self).__init__(
            players, turns=float("inf"),
            deterministic_cache=deterministic_cache)
        self.prob_end = prob_end

    def build_matches(self, cache_mutable=False, noise=0):
        """Build the matches but with cache_mutable False"""
        return super(ProbEndRoundRobin, self).build_matches(False, noise)

    def build_single_match(self, pair, cache_mutable=False, noise=0):
        """Create a single match for a given pair"""
        return Match(pair, self.sample_length(self.prob_end),
                     self.deterministic_cache, cache_mutable, noise)

    def sample_length(self, prob_end):
        """
        Sample length of a game.

        This is using inverse random sample on a probability density function
        <https://en.wikipedia.org/wiki/Probability_density_function> given by:

        f(n) = p_end * (1 - p_end) ^ (n - 1)

        (So the probability of length n is given by f(n))

        Which gives cumulative distribution function
        <https://en.wikipedia.org/wiki/Cumulative_distribution_function>:

        F(n) = 1 - (1 - p_end) ^ n

        (So the probability of length less than or equal to n is given by F(n))

        Which gives for given x = F(n) (ie the random sample) gives n:

        n = ceil((ln(1-x)/ln(1-p_end)))

        This approach of sampling from a distribution is called inverse
        transform sampling
        <https://en.wikipedia.org/wiki/Inverse_transform_sampling>.

        Note that this corresponds to sampling at the end of every turn whether
        or not the Match ends.
        """
        try:
            x = random.random()
            return int(ceil(log(1 - x) / log(1 - self.prob_end)))
        except ZeroDivisionError:
            return float("inf")
        except ValueError:
            return 1



from axelrod import Actions, Player, init_args, random_choice


class Random(Player):
    """A player who randomly chooses between cooperating and defecting."""

    name = 'Random'
    classifier = {
        'memory_depth': 0,  # Memory-one Four-Vector = (p, p, p, p)
        'stochastic': True,
        'makes_use_of': set(),
        'inspects_source': False,
        'manipulates_source': False,
        'manipulates_state': False
    }

    @init_args
    def __init__(self, p=0.5):
        """
        Parameters
        ----------
        p, float
            The probability to cooperate

        Special Cases
        -------------
        Random(0) is equivalent to Defector
        Random(1) is equivalent to Cooperator
        """
        Player.__init__(self)
        self.p = p

    def strategy(self, opponent):
        return random_choice(self.p)

    def __repr__(self):
        return "%s: %s" % (self.name, round(self.p, 2))



from __future__ import absolute_import

from .alternator import Alternator
from .apavlov import APavlov2006, APavlov2011
from .appeaser import Appeaser
from .averagecopier import AverageCopier, NiceAverageCopier
from .axelrod_first import (Davis, RevisedDowning, Feld, Grofman, Nydegger,
                            Joss, Shubik, Tullock, UnnamedStrategy)
from .axelrod_second import Champion, Eatherley, Tester
from .backstabber import BackStabber, DoubleCrosser
from .calculator import Calculator
from .cooperator import Cooperator, TrickyCooperator
from .cycler import AntiCycler, Cycler, CyclerCCD, CyclerCCCD, CyclerCCCCCD
from .darwin import Darwin
from .defector import Defector, TrickyDefector
from .forgiver import Forgiver, ForgivingTitForTat
from .geller import Geller, GellerCooperator, GellerDefector
from .gambler import Gambler, PSOGambler
from .gobymajority import (GoByMajority,
    GoByMajority10, GoByMajority20, GoByMajority40,
    GoByMajority5,
    HardGoByMajority, HardGoByMajority10, HardGoByMajority20, HardGoByMajority40,
    HardGoByMajority5)
from .grudger import Grudger, ForgetfulGrudger, OppositeGrudger, Aggravater
from .grumpy import Grumpy
from .hunter import (
    DefectorHunter, CooperatorHunter, CycleHunter, AlternatorHunter,
    MathConstantHunter, RandomHunter, EventualCycleHunter)
from .inverse import Inverse
from .lookerup import LookerUp, EvolvedLookerUp
from .mathematicalconstants import Golden, Pi, e
from .memoryone import (
    MemoryOnePlayer, ALLCorALLD, FirmButFair, GTFT, SoftJoss,
    StochasticCooperator, StochasticWSLS, ZDExtort2, ZDExtort2v2, ZDExtort4,
    ZDGen2, ZDGTFT2, ZDSet2, WinStayLoseShift)
from .mindcontrol import MindController, MindWarper, MindBender
from .mindreader import MindReader, ProtectedMindReader, MirrorMindReader
from .oncebitten import OnceBitten, FoolMeOnce, ForgetfulFoolMeOnce, FoolMeForever
from .prober import Prober, Prober2, Prober3, HardProber
from .punisher import Punisher, InversePunisher
from .qlearner import RiskyQLearner, ArrogantQLearner, HesitantQLearner, CautiousQLearner
from .rand import Random
from .retaliate import (
    Retaliate, Retaliate2, Retaliate3, LimitedRetaliate, LimitedRetaliate2,
    LimitedRetaliate3)
from .sequence_player import SequencePlayer, ThueMorse, ThueMorseInverse
from .titfortat import (
    TitForTat, TitFor2Tats, TwoTitsForTat, Bully, SneakyTitForTat,
    SuspiciousTitForTat, AntiTitForTat, HardTitForTat, HardTitFor2Tats,
    OmegaTFT)

# Note: Meta* strategies are handled in .__init__.py

strategies = [
    Aggravater,
    ALLCorALLD,
    Alternator,
    AlternatorHunter,
    AntiCycler,
    AntiTitForTat,
    APavlov2006,
    APavlov2011,
    Appeaser,
    ArrogantQLearner,
    AverageCopier,
    BackStabber,
    Bully,
    Calculator,
    CautiousQLearner,
    Champion,
    Cooperator,
    CooperatorHunter,
    CycleHunter,
    CyclerCCCCCD,
    CyclerCCCD,
    CyclerCCD,
    Darwin,
    Davis,
    Defector,
    DefectorHunter,
    DoubleCrosser,
    Eatherley,
    EventualCycleHunter,
    Feld,
    FirmButFair,
    FoolMeForever,
    FoolMeOnce,
    ForgetfulFoolMeOnce,
    ForgetfulGrudger,
    Forgiver,
    ForgivingTitForTat,
    PSOGambler,
    GTFT,
    Geller,
    GellerCooperator,
    GellerDefector,
    GoByMajority,
    GoByMajority10,
    GoByMajority20,
    GoByMajority40,
    GoByMajority5,
    HardGoByMajority,
    HardGoByMajority10,
    HardGoByMajority20,
    HardGoByMajority40,
    HardGoByMajority5,
    Golden,
    Grofman,
    Grudger,
    Grumpy,
    HardProber,
    HardTitFor2Tats,
    HardTitForTat,
    HesitantQLearner,
    Inverse,
    InversePunisher,
    Joss,
    LimitedRetaliate,
    LimitedRetaliate2,
    LimitedRetaliate3,
    EvolvedLookerUp,
    MathConstantHunter,
    MindBender,
    MindController,
    MindReader,
    MindWarper,
    MirrorMindReader,
    NiceAverageCopier,
    Nydegger,
    OmegaTFT,
    OnceBitten,
    OppositeGrudger,
    Pi,
    Prober,
    Prober2,
    Prober3,
    ProtectedMindReader,
    Punisher,
    Random,
    RandomHunter,
    Retaliate,
    Retaliate2,
    Retaliate3,
    RiskyQLearner,
    Shubik,
    SneakyTitForTat,
    SoftJoss,
    StochasticWSLS,
    SuspiciousTitForTat,
    Tester,
    ThueMorse,
    ThueMorseInverse,
    TitForTat,
    TitFor2Tats,
    TrickyCooperator,
    TrickyDefector,
    Tullock,
    TwoTitsForTat,
    WinStayLoseShift,
    ZDExtort2,
    ZDExtort2v2,
    ZDExtort4,
    ZDGen2,
    ZDGTFT2,
    ZDSet2,
    e,
]



from axelrod import Actions, Player


class Alternator(Player):
    """A player who alternates between cooperating and defecting."""

    name = 'Alternator'
    classifier = {
        'memory_depth': 1,
        'stochastic': False,
        'makes_use_of': set(),
        'inspects_source': False,
        'manipulates_source': False,
        'manipulates_state': False
    }

    def strategy(self, opponent):
        if len(self.history) == 0:
            return Actions.C
        if self.history[-1] == Actions.C:
            return Actions.D
        return Actions.C



from __future__ import absolute_import

import logging
import multiprocessing

from .game import Game
from .result_set import ResultSet
from .tournament_type import RoundRobin, ProbEndRoundRobin


class Tournament(object):
    game = Game()

    def __init__(self, players, tournament_type=RoundRobin, name='axelrod',
                 game=None, turns=200, repetitions=10, processes=None,
                 prebuilt_cache=False, noise=0, with_morality=True):
        """
        Parameters
        ----------
        players : list
            A list of axelrod.Player objects
        tournament_type : class
            A class that must be descended from axelrod.TournamentType
        name : string
            A name for the tournament
        game : axelrod.Game
            The game object used to score the tournament
        turns : integer
            The number of turns per match
        repetitions : integer
            The number of times the round robin should be repeated
        processes : integer
            The number of processes to be used for parallel processing
        prebuilt_cache : boolean
            Whether a cache has been passed in from an external object
        noise : float
            The probability that a player's intended action should be flipped
        with_morality : boolean
            Whether morality metrics should be calculated
        keep_matches : boolean
            Whether interaction results should be included in the output
        """
        self.name = name
        self.turns = turns
        self.noise = noise
        if game is not None:
            self.game = game
        self.players = players
        self.repetitions = repetitions
        self.prebuilt_cache = prebuilt_cache
        self.deterministic_cache = {}
        self.tournament_type = tournament_type(
            players, turns, self.deterministic_cache)
        self._with_morality = with_morality
        self._parallel_repetitions = repetitions
        self._processes = processes
        self._logger = logging.getLogger(__name__)
        self.matches = []

    @property
    def players(self):
        return self._players

    @players.setter
    def players(self, players):
        """Ensure that players are passed the tournament attributes"""
        newplayers = []
        for player in players:
            player.set_tournament_attributes(
                length=self.turns,
                 game=self.game,
                 noise=self.noise)
            newplayers.append(player)
        self._players = newplayers

    def play(self):
        """
        Plays the tournament and passes the results to the ResultSet class

        Returns
        -------
        axelrod.ResultSet
        """
        if self._processes is None:
            self._run_serial_repetitions(self.matches)
        else:
            if self._build_cache_required():
                self._build_cache(self.matches)
            self._run_parallel_repetitions(self.matches)

        self.result_set = self._build_result_set()
        return self.result_set

    def _build_result_set(self):
        """
        Build the result set (used by the play method)

        Returns
        -------
        axelrod.ResultSet
        """
        result_set = ResultSet(
            players=self.players,
            matches=self.matches,
            with_morality=self._with_morality)
        return result_set

    def _build_cache_required(self):
        """
        A boolean to indicate whether it is necessary to build the
        deterministic cache.
        """
        return (
            not self.noise and (
                len(self.deterministic_cache) == 0 or
                not self.prebuilt_cache))

    def _build_cache(self, matches):
        """
        For parallel processing, this runs a single round robin in order to
        build the deterministic cache.

        Parameters
        ----------
        matches : list
            The list of matches to update
        """
        self._logger.debug('Playing first round robin to build cache')
        self._run_single_repetition(matches)
        self._parallel_repetitions -= 1

    def _run_single_repetition(self, matches):
        """
        Runs a single round robin and updates the matches list.
        """
        new_matches = self.tournament_type.build_matches(
            cache_mutable=True, noise=self.noise)
        self._play_matches(new_matches)
        self.matches.append(new_matches)

    def _run_serial_repetitions(self, matches):
        """
        Runs all repetitions of the round robin in serial.

        Parameters
        ----------
        matches : list
            The list of matches per repetition to update with results
        """
        self._logger.debug('Playing %d round robins' % self.repetitions)
        for repetition in range(self.repetitions):
            self._run_single_repetition(matches)
        return True

    def _run_parallel_repetitions(self, matches):
        """
        Run all except the first round robin using parallel processing.

        Parameters
        ----------
        matches : list
            The list of matches per repetition to update with results
        """
        # At first sight, it might seem simpler to use the multiprocessing Pool
        # Class rather than Processes and Queues. However, Pool can only accept
        # target functions which can be pickled and instance methods cannot.
        work_queue = multiprocessing.Queue()
        done_queue = multiprocessing.Queue()
        workers = self._n_workers()

        for repetition in range(self._parallel_repetitions):
            work_queue.put(repetition)

        self._logger.debug(
            'Playing %d round robins with %d parallel processes' %
            (self._parallel_repetitions, workers))
        self._start_workers(workers, work_queue, done_queue)
        self._process_done_queue(workers, done_queue, matches)

        return True

    def _n_workers(self):
        """
        Determines the number of parallel processes to use.

        Returns
        -------
        integer
        """
        if (2 <= self._processes <= multiprocessing.cpu_count()):
            n_workers = self._processes
        else:
            n_workers = multiprocessing.cpu_count()
        return n_workers

    def _start_workers(self, workers, work_queue, done_queue):
        """
        Initiates the sub-processes to carry out parallel processing.

        Parameters
        ----------
        workers : integer
            The number of sub-processes to create
        work_queue : multiprocessing.Queue
            A queue containing an entry for each round robin to be processed
        done_queue : multiprocessing.Queue
            A queue containing the output dictionaries from each round robin
        """
        for worker in range(workers):
            process = multiprocessing.Process(
                target=self._worker, args=(work_queue, done_queue))
            work_queue.put('STOP')
            process.start()
        return True

    def _process_done_queue(self, workers, done_queue, matches):
        """
        Retrieves the matches from the parallel sub-processes

        Parameters
        ----------
        workers : integer
            The number of sub-processes in existence
        done_queue : multiprocessing.Queue
            A queue containing the output dictionaries from each round robin
        matches : list
            The list of matches per repetition to update with results
        """
        stops = 0
        while stops < workers:
            results = done_queue.get()

            if results == 'STOP':
                stops += 1
            else:
                new_matches = self.tournament_type.build_matches(
                    cache_mutable=False, noise=self.noise)
                for index_pair, result in results.items():
                    new_matches[index_pair].result = result

                matches.append(new_matches)
        return True

    def _worker(self, work_queue, done_queue):
        """
        The work for each parallel sub-process to execute.

        Parameters
        ----------
        work_queue : multiprocessing.Queue
            A queue containing an entry for each round robin to be processed
        done_queue : multiprocessing.Queue
            A queue containing the output dictionaries from each round robin
        """
        for repetition in iter(work_queue.get, 'STOP'):
            new_matches = self.tournament_type.build_matches(
                cache_mutable=False, noise=self.noise)
            self._play_matches(new_matches)

            results = {index_pair: match.result for
                       index_pair, match in new_matches.items()}
            done_queue.put(results)
        done_queue.put('STOP')
        return True

    def _play_matches(self, matches):
        """
        Play the supplied matches.

        Parameters
        ----------
        matches : dictionary
            Mapping a tuple of player index numbers to an axelrod Match object

        """
        for match in matches.values():
            match.play()


class ProbEndTournament(Tournament):
    """
    A tournament in which the player don't know the length of a given match
    (currently implemented by setting this to be infinite). The length of a
    match is equivalent to randomly sampling after each round whether or not to
    continue.
    """

    def __init__(self, players, tournament_type=ProbEndRoundRobin,
                 name='axelrod', game=None, prob_end=.5, repetitions=10,
                 processes=None, prebuilt_cache=False, noise=0,
                 with_morality=True):
        """
        Parameters
        ----------
        players : list
            A list of axelrod.Player objects
        tournament_type : class
            A class that must be descended from axelrod.TournamentType
        name : string
            A name for the tournament
        game : axelrod.Game
            The game object used to score the tournament
        prob_end : a float
            The probability of a given match ending
        repetitions : integer
            The number of times the round robin should be repeated
        processes : integer
            The number of processes to be used for parallel processing
        prebuilt_cache : boolean
            Whether a cache has been passed in from an external object
        noise : float
            The probability that a player's intended action should be flipped
        with_morality : boolean
            Whether morality metrics should be calculated
        """
        super(ProbEndTournament, self).__init__(
            players, name=name, game=game, turns=float("inf"),
            repetitions=repetitions, processes=processes,
            prebuilt_cache=prebuilt_cache, noise=noise,
            with_morality=with_morality)

        self.prob_end = prob_end
        self.tournament_type = ProbEndRoundRobin(
            players, prob_end, self.deterministic_cache)

    def _build_cache_required(self):
        """
        A cache is never required (as every Match length can be different)
        """
        return False



from __future__ import absolute_import, unicode_literals, print_function

import os
import cloudpickle as pickle

from .tournament import *
from .plot import *
from .ecosystem import *
from .utils import *


class TournamentManager(object):
    """A class to manage and create tournaments."""

    plot_types = {'boxplot': "Payoffs. ", 'payoff': "Payoffs. ",
                  'winplot': "Wins. ", 'sdvplot': "Std Payoffs. ",
                  'pdplot': "Payoff differences. ", 'lengthplot': "Lengths. "}

    ecoturns = {
            'basic_strategies': 1000,
            'cheating_strategies': 10,
            'ordinary_strategies': 1000,
            'strategies': 10,
        }

    def __init__(self, output_directory, with_ecological,
                 pass_cache=True, load_cache=True, save_cache=False,
                 cache_file='./cache.txt', image_format="svg"):
        self._tournaments = []
        self._ecological_variants = []
        self._logger = logging.getLogger(__name__)
        self._output_directory = output_directory
        self._with_ecological = with_ecological
        self._pass_cache = pass_cache
        self._save_cache = save_cache
        self._cache_file = cache_file
        self._deterministic_cache = {}
        self._cache_valid_for_turns = None
        self._load_cache = False
        self._image_format=image_format

        if load_cache and not save_cache:
            self.load_cache = self._load_cache_from_file(cache_file)

    @staticmethod
    def one_player_per_strategy(strategies):
        return [strategy() for strategy in strategies]

    def add_tournament(self, name, players, game=None, turns=200,
                       repetitions=10, processes=None, noise=0,
                       with_morality=True):
        tournament = Tournament(
            name=name,
            players=players,
            turns=turns,
            repetitions=repetitions,
            processes=processes,
            noise=noise,
            with_morality=with_morality)
        self._tournaments.append(tournament)

    def run_tournaments(self):
        t0 = time.time()
        for tournament in self._tournaments:
            self._run_single_tournament(tournament)
        if self._save_cache and not tournament.noise:
            self._save_cache_to_file(self._deterministic_cache, self._cache_file)
        self._logger.info(timed_message('Finished all tournaments', t0))

    def _run_single_tournament(self, tournament):
        self._logger.info(
                'Starting {} tournament: '.format(tournament.name) + self._tournament_label(tournament)
            )

        t0 = time.time()

        if not tournament.noise and self._pass_cache and self._valid_cache(tournament.turns):
            self._logger.debug('Passing cache with %d entries to %s tournament' %
                            (len(self._deterministic_cache), tournament.name))
            tournament.deterministic_cache = self._deterministic_cache
            if self._load_cache:
                tournament.prebuilt_cache = True
        else:
            self._logger.debug('Cache is not valid for %s tournament' %
                            tournament.name)
        tournament.play()

        self._logger.debug(timed_message('Finished %s tournament' % tournament.name, t0))

        if self._with_ecological:
            ecosystem = Ecosystem(tournament.result_set)
            self.run_ecological_variant(tournament, ecosystem)
        else:
            ecosystem = None

        self._generate_output_files(tournament, ecosystem)
        self._cache_valid_for_turns = tournament.turns

        self._logger.debug('Cache now has %d entries' %
                        len(self._deterministic_cache))

        self._logger.info(
            timed_message('Finished all %s tasks' % tournament.name, t0))

    def _valid_cache(self, turns):
        return ((len(self._deterministic_cache) == 0) or
                (len(self._deterministic_cache) > 0) and
                turns == self._cache_valid_for_turns)

    def run_ecological_variant(self, tournament, ecosystem):
        self._logger.debug(
            'Starting ecological variant of %s' % tournament.name)
        t0 = time.time()
        ecosystem.reproduce(self.ecoturns.get(tournament.name))
        self._logger.debug(
            timed_message('Finished ecological variant of %s' % tournament.name, t0))

    def _generate_output_files(self, tournament, ecosystem=None):
        self._save_csv(tournament)
        self._save_plots(tournament, ecosystem,
                         image_format=self._image_format)

    def _save_csv(self, tournament):
        csv = tournament.result_set.csv()
        file_name = self._output_file_path(
                tournament.name, 'csv')
        with open(file_name, 'w') as f:
            f.write(csv)

    def _save_plots(self, tournament, ecosystem=None, image_format="svg"):
        results = tournament.result_set
        plot = Plot(results)
        if not plot.matplotlib_installed:
            self._logger.error('The matplotlib library is not installed. '
                            'No plots will be produced')
            return
        label = self._tournament_label(tournament)
        for plot_type, name in self.plot_types.items():
            title = name + label
            figure = getattr(plot, plot_type)(title=title)
            file_name = self._output_file_path(
                tournament.name + '_' + plot_type, image_format)
            self._save_plot(figure, file_name)
        if ecosystem is not None:
            title = "Eco. " + label
            figure = plot.stackplot(ecosystem, title=title)
            file_name = self._output_file_path(
                    tournament.name + '_reproduce', image_format)
            self._save_plot(figure, file_name)

    def _tournament_label(self, tournament):
        """A label for the tournament for the corresponding title plots"""
        return "Turns: {}, Repetitions: {}, Strategies: {}.".format(tournament.turns,
                                                       tournament.repetitions,
                                                       len(tournament.players))

    def _output_file_path(self, file_name, file_extension):
        return os.path.join(
            self._output_directory,
            file_name + '.' + file_extension)

    @staticmethod
    def _save_plot(figure, file_name, dpi=400):
        figure.savefig(file_name, bbox_inches='tight', dpi=dpi)
        figure.clf()
        plt.close(figure)

    def _save_cache_to_file(self, cache, file_name):
        self._logger.debug(
            'Saving cache with %d entries to %s' % (len(cache), file_name))
        deterministic_cache = DeterministicCache(
            cache, self._cache_valid_for_turns)
        with open(file_name, 'wb') as io:
            pickle.dump(deterministic_cache, io)
        return True

    def _load_cache_from_file(self, file_name):
        try:
            with open(file_name, 'rb') as io:
                deterministic_cache = pickle.load(io)
            self._deterministic_cache = deterministic_cache.cache
            self._cache_valid_for_turns = deterministic_cache.turns
            self._logger.debug(
                'Loaded cache with %d entries' % len(self._deterministic_cache))
            return True
        except IOError:
            self._logger.debug('Cache file not found. Starting with empty cache')
            return False


class ProbEndTournamentManager(TournamentManager):
    """A class to manage and create probabilistic ending tournaments."""

    ecoturns = {
            'basic_strategies_prob_end': 1000,
            'cheating_strategies_prob_end': 10,
            'ordinary_strategies_prob_end': 1000,
            'strategies_prob_end': 10,
        }

    def add_tournament(self, name, players, game=None, prob_end=.01,
                       repetitions=10, processes=None, noise=0,
                       with_morality=True):
        tournament = ProbEndTournament(
            name=name,
            players=players,
            prob_end=prob_end,
            repetitions=repetitions,
            processes=processes,
            noise=noise,
            with_morality=with_morality)
        self._tournaments.append(tournament)

    def _tournament_label(self, tournament):
        """A label for the tournament for the corresponding title plots"""
        return "Prob end: {}, Repetitions: {}, Strategies: {}.".format(tournament.prob_end,
                                                       tournament.repetitions,
                                                       len(tournament.players))


class DeterministicCache(object):

    def __init__(self, cache, turns):
        self.cache = cache
        self.turns = turns



from axelrod import Actions, Player

C, D = Actions.C, Actions.D

class Grudger(Player):
    """A player starts by cooperating however will defect if at any point the opponent has defected."""

    name = 'Grudger'
    classifier = {
        'memory_depth': float('inf'),  # Long memory
        'stochastic': False,
        'makes_use_of': set(),
        'inspects_source': False,
        'manipulates_source': False,
        'manipulates_state': False
    }

    @staticmethod
    def strategy(opponent):
        """Begins by playing C, then plays D for the remaining rounds if the opponent ever plays D."""
        if opponent.defections:
            return D
        return C


class ForgetfulGrudger(Player):
    """A player starts by cooperating however will defect if at any point the
    opponent has defected, but forgets after mem_length matches."""

    name = 'Forgetful Grudger'
    classifier = {
        'memory_depth': 10,
        'stochastic': False,
        'makes_use_of': set(),
        'inspects_source': False,
        'manipulates_source': False,
        'manipulates_state': False
    }

    def __init__(self):
        """Initialised the player."""
        super(ForgetfulGrudger, self).__init__()
        self.mem_length = 10
        self.grudged = False
        self.grudge_memory = 0

    def strategy(self, opponent):
        """Begins by playing C, then plays D for mem_length rounds if the opponent ever plays D."""
        if self.grudge_memory >= self.mem_length:
            self.grudge_memory = 0
            self.grudged = False

        if self.grudged:
            self.grudge_memory += 1
            return D
        elif D in opponent.history[-1:]:
            self.grudged = True
            return D
        return C

    def reset(self):
        """Resets scores and history."""
        Player.reset(self)
        self.grudged = False
        self.grudge_memory = 0


class OppositeGrudger(Player):
    """A player starts by defecting however will cooperate if at any point the opponent has cooperated."""

    name = 'Opposite Grudger'
    classifier = {
        'memory_depth': float('inf'),  # Long memory
        'stochastic': False,
        'makes_use_of': set(),
        'inspects_source': False,
        'manipulates_source': False,
        'manipulates_state': False
    }

    @staticmethod
    def strategy(opponent):
        """Begins by playing D, then plays C for the remaining rounds if the opponent ever plays C."""
        if opponent.cooperations:
            return C
        return D


class Aggravater(Player):
    """Grudger, except that it defects on the first 3 turns"""

    name = 'Aggravater'
    classifier = {
        'memory_depth': float('inf'),  # Long memory
        'stochastic': False,
        'makes_use_of': set(),
        'inspects_source': False,
        'manipulates_source': False,
        'manipulates_state': False
    }

    @staticmethod
    def strategy(opponent):
        if len(opponent.history) < 3:
            return D
        elif opponent.defections:
            return D
        return C



from __future__ import absolute_import

# The order of imports matters!
from .actions import Actions, flip_action
from .random_ import random_choice
from .plot import Plot
from .game import DefaultGame, Game
from .player import init_args, is_basic, obey_axelrod, update_history, Player
from .mock_player import MockPlayer, simulate_play
from .match import Match
from .strategies import *
from .tournament_type import *
from .tournament import Tournament, ProbEndTournament
from .tournament_manager import TournamentManager, ProbEndTournamentManager
from .tournament_manager_factory import (TournamentManagerFactory,
                                         ProbEndTournamentManagerFactory)
from .result_set import ResultSet
from .ecosystem import Ecosystem
from .utils import run_tournaments, run_prob_end_tournaments, setup_logging



from ..player import is_basic, obey_axelrod
from ._strategies import *

# `from ._strategies import *` import the collection `strategies`
# Now import the Meta strategies. This cannot be done in _strategies
# because it creates circular dependencies

from .meta import (
    MetaPlayer, MetaMajority, MetaMinority, MetaWinner, MetaHunter,
    MetaMajorityMemoryOne, MetaWinnerMemoryOne, MetaMajorityFiniteMemory,
    MetaWinnerFiniteMemory, MetaMajorityLongMemory, MetaWinnerLongMemory,
    MetaMixer
    )

strategies.extend((MetaHunter, MetaMajority, MetaMinority, MetaWinner,
                   MetaMajorityMemoryOne, MetaWinnerMemoryOne,
                   MetaMajorityFiniteMemory, MetaWinnerFiniteMemory,
                   MetaMajorityLongMemory, MetaWinnerLongMemory, MetaMixer))

# Distinguished strategy collections in addition to
# `strategies` from _strategies.py

demo_strategies = [Cooperator, Defector, TitForTat, Grudger, Random]
basic_strategies = [s for s in strategies if is_basic(s())]
ordinary_strategies = [s for s in strategies if obey_axelrod(s())]
cheating_strategies = [s for s in strategies if not obey_axelrod(s())]



from axelrod import Actions, Player, init_args, flip_action

C, D = Actions.C, Actions.D


class TitForTat(Player):
    """
    A player starts by cooperating and then mimics the previous action of the
    opponent.

    Note that the code for this strategy is written in a fairly verbose
    way. This is done so that it can serve as an example strategy for
    those who might be new to Python.
    """

    # These are various properties for the strategy
    name = 'Tit For Tat'
    classifier = {
        'memory_depth': 1,  # Four-Vector = (1.,0.,1.,0.)
        'stochastic': False,
        'makes_use_of': set(),
        'inspects_source': False,
        'manipulates_source': False,
        'manipulates_state': False
    }

    def strategy(self, opponent):
        """This is the actual strategy"""
        # First move
        if len(self.history) == 0:
            return C
        # React to the opponent's last move
        if opponent.history[-1] == D:
            return D
        return C


class TitFor2Tats(Player):
    """A player starts by cooperating and then defects only after two defects by opponent."""

    name = "Tit For 2 Tats"
    classifier = {
        'memory_depth': 2,  # Long memory, memory-2
        'stochastic': False,
        'makes_use_of': set(),
        'inspects_source': False,
        'manipulates_source': False,
        'manipulates_state': False
    }

    @staticmethod
    def strategy(opponent):
        return D if opponent.history[-2:] == [D, D] else C


class TwoTitsForTat(Player):
    """A player starts by cooperating and replies to each defect by two defections."""

    name = "Two Tits For Tat"
    classifier = {
        'memory_depth': 2,  # Long memory, memory-2
        'stochastic': False,
        'makes_use_of': set(),
        'inspects_source': False,
        'manipulates_source': False,
        'manipulates_state': False
    }

    @staticmethod
    def strategy(opponent):
        return D if D in opponent.history[-2:] else C


class Bully(Player):
    """A player that behaves opposite to Tit For Tat, including first move.

    Starts by defecting and then does the opposite of opponent's previous move.
    This is the complete opposite of TIT FOR TAT, also called BULLY in literature.
    """

    name = "Bully"
    classifier = {
        'memory_depth': 1,   # Four-Vector = (0, 1, 0, 1)
        'stochastic': False,
        'makes_use_of': set(),
        'inspects_source': False,
        'manipulates_source': False,
        'manipulates_state': False
    }

    @staticmethod
    def strategy(opponent):
        return C if opponent.history[-1:] == [D] else D


class SneakyTitForTat(Player):
    """Tries defecting once and repents if punished."""

    name = "Sneaky Tit For Tat"
    classifier = {
        'memory_depth': float('inf'),  # Long memory
        'stochastic': False,
        'makes_use_of': set(),
        'inspects_source': False,
        'manipulates_source': False,
        'manipulates_state': False
    }

    def strategy(self, opponent):
        if len(self.history) < 2:
            return "C"
        if D not in opponent.history:
            return D
        if opponent.history[-1] == D and self.history[-2] == D:
            return "C"
        return opponent.history[-1]


class SuspiciousTitForTat(Player):
    """A TFT that initially defects."""

    name = "Suspicious Tit For Tat"
    classifier = {
        'memory_depth': 1, # Four-Vector = (1.,0.,1.,0.)
        'stochastic': False,
        'makes_use_of': set(),
        'inspects_source': False,
        'manipulates_source': False,
        'manipulates_state': False
    }

    @staticmethod
    def strategy(opponent):
        return C if opponent.history[-1:] == [C] else D


class AntiTitForTat(Player):
    """A strategy that plays the opposite of the opponents previous move.
    This is similar to BULLY above, except that the first move is cooperation."""

    name = 'Anti Tit For Tat'
    classifier = {
        'memory_depth': 1,  # Four-Vector = (1.,0.,1.,0.)
        'stochastic': False,
        'makes_use_of': set(),
        'inspects_source': False,
        'manipulates_source': False,
        'manipulates_state': False
    }

    @staticmethod
    def strategy(opponent):
        return D if opponent.history[-1:] == [C] else C


class HardTitForTat(Player):
    """A variant of Tit For Tat that uses a longer history for retaliation."""

    name = 'Hard Tit For Tat'
    classifier = {
        'memory_depth': 3,  # memory-three
        'stochastic': False,
        'makes_use_of': set(),
        'inspects_source': False,
        'manipulates_source': False,
        'manipulates_state': False
    }

    @staticmethod
    def strategy(opponent):
        # Cooperate on the first move
        if not opponent.history:
            return C
        # Defects if D in the opponent's last three moves
        if D in opponent.history[-3:]:
            return D
        # Otherwise cooperates
        return C


class HardTitFor2Tats(Player):
    """A variant of Tit For Two Tats that uses a longer history for
    retaliation."""

    name = "Hard Tit For 2 Tats"
    classifier = {
        'memory_depth': 3,  # memory-three
        'stochastic': False,
        'makes_use_of': set(),
        'inspects_source': False,
        'manipulates_source': False,
        'manipulates_state': False
    }

    @staticmethod
    def strategy(opponent):
        # Cooperate on the first move
        if not opponent.history:
            return C
        # Defects if two consecutive D in the opponent's last three moves
        history_string = "".join(opponent.history[-3:])
        if 'DD' in history_string:
            return D
        # Otherwise cooperates
        return C


class OmegaTFT(Player):
    """OmegaTFT modifies TFT in two ways:
       -- checks for deadlock loops of alternating rounds of (C, D) and (D, C),
       and attempting to break them
       -- uses a more sophisticated retaliation mechanism that is noise tolerant.
    """

    name = "Omega TFT"
    classifier = {
        'memory_depth': float('inf'),
        'stochastic': False,
        'makes_use_of': set(),
        'inspects_source': False,
        'manipulates_source': False,
        'manipulates_state': False
    }

    @init_args
    def __init__(self, deadlock_threshold=3, randomness_threshold=8):
        Player.__init__(self)
        self.deadlock_threshold = deadlock_threshold
        self.randomness_threshold = randomness_threshold
        self.randomness_counter = 0
        self.deadlock_counter = 0

    def strategy(self, opponent):
        # Cooperate on the first move
        if len(self.history) == 0:
            return C
        # TFT on round 2
        if len(self.history) == 1:
            return D if opponent.history[-1:] == [D] else C

        # Are we deadlocked? (in a CD -> DC loop)
        if (self.deadlock_counter >= self.deadlock_threshold):
            self.move = C
            if self.deadlock_counter == self.deadlock_threshold:
                self.deadlock_counter = self.deadlock_threshold + 1
            else:
                self.deadlock_counter = 0
        else:
            # Update counters
            if opponent.history[-2:] == [C, C]:
                self.randomness_counter -= 1
            # If the opponent's move changed, increase the counter
            if opponent.history[-2] != opponent.history[-1]:
                self.randomness_counter += 1
            # If the opponent's last move differed from mine, increase the counter
            if self.history[-1] == opponent.history[-1]:
                self.randomness_counter+= 1
            # Compare counts to thresholds
            # If randomness_counter exceeds Y, Defect for the remainder
            if self.randomness_counter >= 8:
                self.move = D
            else:
                # TFT
                self.move = D if opponent.history[-1:] == [D] else C
                # Check for deadlock
                if opponent.history[-2] != opponent.history[-1]:
                    self.deadlock_counter += 1
                else:
                    self.deadlock_counter = 0
        return self.move

    def reset(self):
        Player.reset(self)
        self.randomness_counter = 0
        self.deadlock_counter = 0



from __future__ import absolute_import

import time
import logging

from .tournament_manager_factory import (TournamentManagerFactory,
                                         ProbEndTournamentManagerFactory)


def timed_message(message, start_time):
    elapsed_time = time.time() - start_time
    return message + " in %.1fs" % elapsed_time


def setup_logging(logging_destination='console', verbosity='INFO'):
    """Sets up logging. Call this outside of run_tournaments to avoid
    accumulating logging handlers."""
    logHandlers = {
        'console': logging.StreamHandler,
        'none': logging.NullHandler,
    }
    if logging_destination == 'file':
        logHandler = logging.FileHandler('./axelrod.log')
    else:
        logHandler = logHandlers[logging_destination]()

    logFormatters = {
        'console': '%(message)s',
        'none': '',
        'file': '%(asctime)s - %(name)s - %(levelname)s - %(message)s'
    }
    logFormatter = logging.Formatter(logFormatters[logging_destination])

    logHandler.setFormatter(logFormatter)
    logger = logging.getLogger('axelrod')
    logger.setLevel(verbosity.upper())
    logger.addHandler(logHandler)


def build_exclusions_dict(exclude_basic, exclude_ordinary,
                          exclude_cheating, exclude_combined):
    """A utility function to return a dictionary mapping tournament string names
    to booleans."""
    return {
        'basic_strategies': exclude_basic,
        'ordinary_strategies': exclude_ordinary,
        'cheating_strategies': exclude_cheating,
        'strategies': exclude_combined}


def run_tournaments(cache_file='./cache.txt',
                    output_directory='./',
                    repetitions=10,
                    turns=200,
                    processes=None,
                    no_ecological=False,
                    rebuild_cache=False,
                    exclude_combined=False,
                    exclude_basic=False,
                    exclude_cheating=False,
                    exclude_ordinary=False,
                    noise=0,
                    image_format="svg"):

    exclusions_dict = build_exclusions_dict(exclude_basic, exclude_ordinary,
                                            exclude_cheating, exclude_combined)

    exclusions = [key for key, value in exclusions_dict.items() if value]

    manager = TournamentManagerFactory.create_tournament_manager(
        output_directory=output_directory,
        no_ecological=no_ecological,
        rebuild_cache=rebuild_cache,
        cache_file=cache_file,
        exclusions=exclusions,
        processes=processes,
        turns=turns,
        repetitions=repetitions,
        noise=noise,
        image_format=image_format)

    manager.run_tournaments()


def run_prob_end_tournaments(cache_file='./cache.txt',
                    output_directory='./',
                    repetitions=10,
                    prob_end=.01,  # By default have mean of 100 rounds
                    processes=None,
                    no_ecological=False,
                    rebuild_cache=False,
                    exclude_combined=False,
                    exclude_basic=False,
                    exclude_cheating=False,
                    exclude_ordinary=False,
                    noise=0,
                    image_format="svg"):

    exclusions_dict = build_exclusions_dict(exclude_basic, exclude_ordinary,
                                            exclude_cheating, exclude_combined)

    exclusions = [key for key, value in exclusions_dict.items() if value]

    manager = ProbEndTournamentManagerFactory.create_tournament_manager(
        output_directory=output_directory,
        no_ecological=no_ecological,
        rebuild_cache=rebuild_cache,
        cache_file=cache_file,
        exclusions=exclusions,
        processes=processes,
        prob_end=prob_end,
        repetitions=repetitions,
        noise=noise,
        image_format=image_format)

    manager.run_tournaments()
